function [input,finalOutput] = reSynth(inFile,fL,nd,nud,overlap,outputFile, pitchFactor)
% function [input,finalOutput] =
% reSynth(inFile,fL,nd,nud,overlap,outputFile,pitchFactor)
% takes input file, performs frame by frame spetral analysis
% and resynthesises using sine oscillators and noise filtered
% by a bank of parametric equalisers. 
% fL is the frame length in samples
% nd is the number of decimated levels in the wavelet analysis
% nud is the number of undecimated levels
% overlap is the overlap factor for th analysis
% output file is name of wav file output
% pitchFactor determines pitch scaling of output
% (1 = unchanged pitch)

if (fL ~= 513&fL~=1025)
    % do not continue if frame length is no 513 or 1025
    error('frame length can only be 513 or 1025');
end
if (fL==1025 & nd+nud>10)
    % do not continue if number of scales is too high
    error('maximum of 10 levels for wavelet analysis with a frame length of 1025');
end
if (fL==513 & nd+nud>9)
    % do not continue if number of scales is too high
    error('maximum of 9 levels for wavelet analysis with a frame length of 513');
end

global frameCount;
frameCount = 1;
global frameLength;
frameLength = fL;
global Fs;
Fs = 44100;
bits = 16;
global freqMult;
freqMult = (2*pi)/Fs;
global ampScaler;
ampScaler = 1/256;   % amp scaler for 1025 point Hann window
global synLength;
synLength = floor(frameLength/overlap);
global synProp;
synProp = synLength/frameLength;    % length of synthesis frame as proportion of analysis frame
zeroPad = 8;
range = 2;
chiSquaredThreshold = 0.006; % see chapter 4 of thesis
global iterations;
iterations = 1; % number of times around da and df estimation loop
global meanMult;
meanMult = 20/log(10); % constant for calculating mean amplitude of sinusoids
global daPointerMult;
daPointerMult = 99/96;  % constant for estimating da
global dfPointerMult;
dfPointerMult = 99/260;
global temp;
temp = zeros(1,synLength);
residualRange = 32; % number of bins above and below peak that are zeroed for residual
global sqrt0_5;
sqrt0_5 = sqrt(0.5);

% load da, df and chiSquared data
[daX,daY,dfX,dfY,chiExpect,ampCorrect] = tableLoad();
% generate windows
global Hn;
global tHn;
global fHn;
windowGenerate();
% calculate zero-padded FFT size and allocate zero-padded input frame
global FFTSize;
FFTSize = 2^(floor(log2(frameLength * zeroPad)));
global complexSize;
complexSize = (FFTSize*0.5)+1;
global standardFFT;
standardFFT = zeros(1,complexSize);
global freqFFT;
freqFFT = zeros(1,complexSize);
global timeFFT;
timeFFT = zeros(1,complexSize);
global tempFFT;
tempFFT =zeros(1,FFTSize); % temporary storage for FFT data
global zeroPadded;
zeroPadded = zeros(1,FFTSize);  
global fL_5;
fL_5 = floor(frameLength * 0.5);
global binWidth;
binWidth = 44100/FFTSize;
global recipBinWidth;
recipBinWidth = 1/binWidth;
global freqLinkThreshold;
freqLinkThreshold = 100;  
global ampFadeMult;
ampFadeMult = ((10/log(10))*(10^2.4-10^-2.4)/48)/(10^(48/20)); 
global A;   % matrix for parabola fitting
A = ones(3,3);
global B;   % vector for parabola fitting
B = ones(3,1);
global SineTable;
sineTable = sin(((2*pi)/8192)*[0:8191]);
global lnTable;
lnTable = log([1:8192]/8192);
global tanTable;
tanTable = tan((pi/8192)*[0:8191]);
global tanLookupMult;
tanLookupMult = 8192/(2*pi);
% set up FFT algorithm
fftw('planner','exhaustive');   

% allocate analysis arrays
amp = zeros(1,complexSize);
phi = zeros(1,complexSize);
freq = zeros(1,complexSize);
da = zeros(1,complexSize);
df = zeros(1,complexSize);
chiSquared = -1*ones(1,complexSize);
endFreq = zeros(0,0);
endAmp = zeros(0,0);
endPhi = zeros(0,0);
sineDf = zeros(0,0);

% initialise wavelet analysis filters etc.
global order;
order = 3;
global LPF; % scaling filter
global HPF; % wavelet filter
[LPF,HPF] = BSplineMake(order);
global lH;
lH = length(HPF);
global lL;
lL = length(LPF);
global INIT; % Fourier domain wavelet initialisation filter 
INIT = sinc(linspace(0,0.5,complexSize)).^(order+1);
global num_dec;
num_dec = nd;
global num_undec;
num_undec = nud;
global p;   % number of holes in decimated wavelet filters
p = ceil(2^(num_undec-1));    % ceil in case num_undec==0
global p2; % number of holes in decimated split wavelet filters
p2 = p*2;
global decSamples;  % number of sample blocks in each decimated level
decSamples = 2.^[1:num_dec];
if (num_undec==0)
    decSamples = decSamples * 0.5;   % when only decimated the first level is still undecimated
end
global Fs2Powers;
Fs2Powers = Fs*(0.5.^[0:num_dec+num_undec-1]);
global decFreqMult; % vector of multipliers for deriving decimated wavelet frequencies
decFreqMult = 0.5.^[1:num_dec];
if (num_undec==0)
    decFreqMult = decFreqMult*2;    
end
global Fs2Pi;
Fs2Pi = Fs/(2*pi);
global decLen;  % lengths of output sequences of decimated wavelets
decLen = floor(synLength*decFreqMult);
global splitFreqLimit;
splitFreqLimit = (0.4678-0.2919)*Fs2Powers; % maximum split width allowed
global scaleFreqRange;
scaleFreqRange = [Fs*0.5,(0.2829)*Fs2Powers]; % frequency ranges for each scale
centreFreqMag = 0.409177;  
global centreScaleFreqs;
centreScaleFreqs = centreFreqMag*Fs2Powers;
centreFreqMag = ((sin(2*pi*centreFreqMag)^2)/(centreFreqMag*(centreFreqMag-0.5)^2))^(order+1);

global decMag;  % initialised in analyticWaveletComb
decMag = cell(num_undec);
global decFreq;
decFreq = cell(num_undec);
global decBW;
decBW = cell(num_undec);
global undecMag;
undecMag = zeros(num_undec,synLength+1);
global undecFreq;
undecFreq = zeros(num_undec,synLength);
global undecBW;
undecBW = zeros(num_undec,synLength);

global noise;   % vector containing Gaussian noise
noise = zeros(1,synLength);
global noiseOut;
noiseOut = zeros(1,synLength); % vector containing equalised noise
global rand1;
rand1 = zeros(1,synLength);
global rand2;
rand2 = zeros(1,synLength);
global noiseVar;
noiseVar = 0.25; % variance of Gaussian noise

global xnm1;    % store previous values of x and y for EQ algorithm
xnm1 = zeros(1,num_dec+num_undec);
global xnm2;
xnm2 = zeros(1,num_dec+num_undec);
global ynm1;
ynm1 = zeros(1,num_dec+num_undec);
global ynm2;
ynm2 = zeros(1,num_dec+num_undec);
global dB96Mult;
dB96Mult = 2*10^(96/20);    % multiply by 2 to give normalised wavelet output magnitude

% EQ coefficient pre-calculations
global G0;
G0 = 1; % DC gain of EQ
global G;
G = dB96Mult; % peak gain of EQ
global GB;
GB = sqrt0_5*dB96Mult;   % BW gain of EQ
global G2;
G2 = G^2;
global GB2;
GB2 = GB^2;
global G02;
G02 = G0^2;
global G2_GB2;
G2_GB2 = G2-GB2;
global GB2_G02;
GB2_G02 = GB2-G02;
global GB2G02G2;
GB2G02G2 = GB2_G02/G2_GB2;
global PI2;
PI2 = pi^2;
global G2PI2;
G2PI2 = G2*PI2;
global G2_G02;
G2_G02 = G2-G02;

sampleCount = 1;
% find length of first file and number of channels
params = wavread(inFile,'size');
input = wavread(inFile);    
numSamples = params(1,1);
numChannels = params(1,2);
if numChannels ~= 1
    % do not continue if input is not mono
    error('only mono files can be processed');
end

% ensure finalOutput matrix will be big enough to accomodate a whole number of
% frames
outSamples = (ceil(numSamples/frameLength) + 1) * frameLength;
% pre-allocate memory for finalOutput, input frames, finalOutput frames
finalOutput = zeros(numChannels,outSamples);
frame = zeros(numChannels,frameLength);
output = zeros(numChannels,frameLength);

while sampleCount < numSamples
    outPointer = sampleCount + (frameLength-1);
    if outPointer > numSamples
        outPointer = numSamples;
    end
    sampleRange = [sampleCount, outPointer];
    % read in a single frame
    [frame] = rot90(wavread(inFile, sampleRange));    % rotate from column to row
    if length(frame) < frameLength
        % at end of file pad the remainder with zeros to produce a full 
        % length analysis frame
        % calculate difference between input and frame length
        diff = frameLength - length(frame);
        frame = [frame, zeros(1,diff)];   %append zeros
    end
    doFFT(frame,Hn,tHn,fHn,complexSize);
    amp = abs(standardFFT);
    %sineIndex = findPeaks(amp,complexSize);    % MATLAB version of function
    sineIndex = CFindPeaks(amp,complexSize);    % MEX version of function
    [freq,sineIndex] = estimateFreqs(sineIndex,freq);
    % MATLAB version of function
    %[df,da,chiSquared] = timeOffsetFit(freq,sineIndex,range,df,da,chiSquared,binWidth);
    % C version of function
    CTimeOffsetFit(freq,sineIndex,range,df,da,chiSquared,binWidth,standardFFT,timeFFT);
    if (fL==513)
        df = 4.0*df;    % scale offset measures to frame length if not 1025
        da = 2.0*da;
        chiSquared = chiSquared.*0.25;
    end
    % MATLAB version of function
    %[df,da,chiSquared,sineIndex] = dEstimate(df,da,daX,daY,dfX,dfY,chiSquared,chiExpect,sineIndex,chiSquaredThreshold);
    % MEX version of function
    [sineIndex] = CDEstimate(df,da,daX,daY,flipud(rot90(dfX)),dfY,chiSquared,flipud(rot90(chiExpect)),sineIndex,chiSquaredThreshold,iterations);
    if (fL==513)
        df = 2.0.*df;  % df must be modified where frame length is ~= 1025
    end
    [correctFreq] = actualFreqMean(freq,df,da,sineIndex);
    [sineIndex,freq] = freqTest(freq,correctFreq,sineIndex);
    if (fL==513)
        df = 0.5*df;     % switch back for amp correction
    end
    [synAmp] = ampCorrection(amp,sineIndex,ampCorrect,dfY,daY,df,da,freq);
    if(fL==513)
        df = 2.0*df;    % and back again for synthesis
    end


    prevEndFreq = endFreq;  % transfer previous frame's data
    prevDf = sineDf;
    prevEndAmp = endAmp;
    prevEndPhi = endPhi;
    % produce sinusoidal synthesis parameters
    [synAmp,df,da,endFreq,sineDf,linkIndex,linkDistance,newIndex] = sineLink(sineIndex,freq,synAmp,df,da,prevEndFreq);
 
    % scale pitch of sinusoids
    if (pitchFactor~=1.0)
        freq = freq * pitchFactor;
        df = df * pitchFactor;
    end
    % synthesize starting, continued and finishing sinusoids from previous
    % frame
    [unLinkedSine,continuedSine,newSine,endAmp,endPhi] = synth(sineIndex,linkIndex,newIndex,prevEndFreq,prevEndAmp,prevEndPhi,prevDf,freq,synAmp,df,da);
    
    % MATLAB version of function
    %[standardFFT] = spectralSubtract(standardFFT,sineIndex,residualRange,complexSize,recipBinWidth,df);
    % MEX version of function
    CSpectralSubtract(standardFFT,sineIndex,residualRange,complexSize,recipBinWidth,df);
    % do wavelet analysis
    [decMag,decFreq,decBW,undecMag,undecFreq,undecBW] = AnalyticWaveletCombSplit(standardFFT,tempFFT,FFTSize,complexSize,frameLength,num_undec,num_dec,INIT,fL_5,decFreqMult,Fs2Pi,Fs,Fs2Powers,decMag,decFreq,decBW,undecMag,undecFreq,undecBW,scaleFreqRange,splitFreqLimit,centreScaleFreqs);
    [decMag,undecMag] = waveletMagCorrect(num_dec,num_undec,decFreq,decMag,undecFreq,undecMag,centreFreqMag,1/Fs,order);
    
    % produce random sequences for noise
    rand1 = rand(1,synLength);
    rand2 = rand(1,synLength);
    % reduce noise to -96 dB before EQ boosting
    noise =  sineTable(ceil(rand1*8192)).*sqrt(-2*noiseVar*lnTable(ceil(rand2*8192)));
    rms = sqrt(sum((noise).^2)/length(noise)); 
    noise = (1.584893192461114e-005/rms)*noise;
    
    % scale pitch of residual
    if (pitchFactor~=1.0)
        for n = 1:length(decFreq)
            decFreq{n} = decFreq{n}*pitchFactor;
        end
        undecFreq = undecFreq*pitchFactor;
    end
    % MATLAB version of function
    %EQ();
    % MEX version of function
    [noiseOut] = CEQ(noise,decMag,decFreq,decBW,flipud(rot90(undecMag)),flipud(rot90(undecFreq)),flipud(rot90(undecBW)),num_dec,num_undec,xnm1,xnm2,ynm1,ynm2,synLength,Fs,decLen,decSamples); 
    
    % finished so clear arrays
    amp = zeros(1,complexSize);
    phi = zeros(1,complexSize);
    freq = zeros(1,complexSize);
    da = zeros(1,complexSize);
    df = zeros(1,complexSize);
    chiSquared = -1*ones(1,complexSize);
    lastSineIndex = sineIndex;
    clear sineIndex;

    frameCount = frameCount+1;
    
    sineOutput = unLinkedSine+continuedSine+newSine;
    %place outputs in single matrix to be written to .wav file  
    finalOutput(sampleCount:(sampleCount + (synLength-1))) =  finalOutput(sampleCount:(sampleCount + (synLength-1))) + noiseOut + sineOutput;
    sampleCount = sampleCount + synLength; 
end
if (max(abs(finalOutput>0)))
    if(max(abs(finalOutput>1)))
        %finalOutput = finalOutput/max(abs(finalOutput));    % to prevent clipping
    end
    wavwrite(rot90(finalOutput,-1),Fs,bits,outputFile); % rotate back to column for wav write
end
% play the finalOutput 
wavplay(finalOutput, Fs);
%**************************************************************************
function[daX,daY,dfX,dfY,chiExpect,ampCorrect] = tableLoad();
% table to convert daMeasure to dB
daY = linspace(0,96,100);
load daTable0_96_100;   % load data for da

% table to convert dfMeasure to Hz
dfY = linspace(0,260,100);
load dfTable0_260_100;   % load data for df

% load chiSquared table
load chiSquaredTable0_96_260
% load amp correction table 
load ampTable0_96_260;
%**************************************************************************
function[] = windowGenerate();
global Hn;
global tHn;
global fHn;
global frameLength;
arg = linspace(-1,1,frameLength);
% Hann window
Hn = 0.5 + (0.5*(cos(pi.*arg)));

% time-ramped Hann window
range = (frameLength-1)/2;
tHn = Hn .* [-range:range];
% frequency ramped window
fHn = 0.5*-sin(pi.*arg)*(2*pi/frameLength);
%**************************************************************************
function[] = doFFT(frame,Hn,tHn,fHn,complexSize);
global standardFFT;
global timeFFT;
global freqFFT;
global FFTSize;
global frameLength;
global zeroPadded;
zeroPadded = zeros(1,FFTSize);  % initialise
global fL_5;
global tempFFT;
% standard FFT
windowed = Hn .* frame;
zeroPadded(1:fL_5+1) = windowed(fL_5+1:frameLength);
zeroPadded(FFTSize-fL_5+1:FFTSize) = windowed(1:fL_5);
tempFFT = fft(zeroPadded);
standardFFT = tempFFT(1:complexSize);
% time ramped FFT
windowed = tHn .* frame;
zeroPadded(1:fL_5+1) = windowed(fL_5+1:frameLength);
zeroPadded(FFTSize-fL_5+1:FFTSize) = windowed(1:fL_5);
tempFFT = fft(zeroPadded);
timeFFT = tempFFT(1:complexSize);
% frequency ramped FFT
windowed = fHn .* frame;
zeroPadded(1:fL_5+1) = windowed(fL_5+1:frameLength);
zeroPadded(FFTSize-fL_5+1:FFTSize) = windowed(1:fL_5);
tempFFT = fft(zeroPadded);
freqFFT = tempFFT(1:complexSize);
%**************************************************************************
function[sineIndex] = findPeaks(amp,complexSize);
a = 1;
upper = complexSize - 4;
for n = 5:upper
    if ((sum(amp(n-4:n-1)<amp(n))==4) & (sum(amp(n+1:n+4)<amp(n))==4))
        sineIndex(a) = n; % could be a sine
        a = a+1;
    end
end
if(a==1)    % if no peaks return empty vector
    sineIndex = zeros(0,0);
end
%**************************************************************************
function[freq,sineIndex] = estimateFreqs(sineIndex,freq);
global FFTSize
global binWidth;
global standardFFT;
global freqFFT;
limit = 71/binWidth;
% perform complex division and multiply imaginary part by FFTsize/2Pi
offsets = ((FFTSize*0.5)/pi) * imag(freqFFT(sineIndex)./standardFFT(sineIndex));

% calculate frequency in Hz
index = find(abs(offsets)<limit);
sineIndex = sineIndex(index);   % only those peaks within frequency range are sines
temp = sineIndex-1-offsets(index);
freq(sineIndex) = temp * binWidth;
f = freq(sineIndex);
%**************************************************************************
function[df,da,chiSquared] = timeOffsetFit(freq,sineIndex,range,df,da,chiSquared,binWidth);
global standardFFT;
global timeFFT;
global A;
global B;
x = zeros(1,(2*range)+1);
y = zeros(1,(2*range)+1);

for n = 1:length(sineIndex)
    ix = sineIndex(n);
    lowerRange = ix - range;
    upperRange = ix + range;
    x(1:upperRange-lowerRange+1) = binWidth*[lowerRange-1:upperRange-1] - freq(ix);
    y(1:upperRange-lowerRange+1) = real(timeFFT(lowerRange:upperRange)./standardFFT(lowerRange:upperRange));
    sx = sum(x);
    sx2 = sum(x.^2);
    sx3 = sum(x.^3);
    A(1,1) = length(x);
    A(1,2) = sx;
    A(1,3) = sx2;
    A(2,1) = sx;
    A(2,2) = sx2;
    A(2,3) = sx3;
    A(3,1) = sx2;
    A(3,2) = sx3;
    A(3,3) = sum(x.^4);
    B(1) = sum(y);
    B(2) = sum(x.*y);
    B(3) = sum((x.^2).*y);
    roots = rot90(A\B,-1);
    % calculate sum of squared residuals for variance test
    df(ix) = roots(2);
    da(ix) = roots(3);
    residual = (y - polyval(roots,x)).^2;
    chiSquared(ix) = sum(residual);
end
%**************************************************************************
function[df,da,chiSquared,sineIndex] = dEstimate(df,da,daX,daY,dfX,dfY,chiSquared,chiExpect,sineIndex,chiSquaredThreshold);
global iterations;
dfEst = zeros(1,iterations+1);
daEst = zeros(1,iterations+1);
global daPointerMult;
global dfPointerMult;
for n = 1:length(sineIndex)
        ix = sineIndex(n);
        % find absolute value and store polarity of da and df measures
        dfPol = abs(df(ix))/df(ix);
        daPol = abs(da(ix))/da(ix);
        df(ix) = abs(df(ix));
        da(ix) = abs(da(ix));
        % find maximum
        xMax = max(daX(:,1));
        xMin = min(daX(:,1));
        if (xMax < da(ix))
            da(ix) = xMax;
        end
        if (da(ix) < xMin)
            da(ix) = xMin;
        end
        daEst(1) = interp1DLin(daX(:,1),daY,da(ix));
        if(daEst(1) < 0.0)
            daEst(1) = 0.0;
        end
        if(daEst(1) > 96.0)
            daEst(1) = 96.0;
        end
                      
        daPointer = 1 + (daEst(1) * daPointerMult);
        if (daPointer < 1)
            daPointer = 1;
        end
        if (daPointer > 100)
            daPointer = 100;
        end
        % find maximum
        xMax = max(dfX(ceil(daPointer),:));
        xMin = min(dfX(ceil(daPointer),:));
        temp = df(ix);
        if (xMax < temp)
            temp = xMax;
        end
        if (temp < xMin)
            temp = xMin;
        end
        hi = interp1DLin(dfX(ceil(daPointer),:),dfY,temp);
        xMax = max(dfX(floor(daPointer),:));
        xMin = min(dfX(floor(daPointer),:));
        temp = df(ix);
        if (xMax < temp)
            temp = xMax;
        end
        if (temp < xMin)
            temp = xMin;
        end
        lo = interp1DLin(dfX(floor(daPointer),:),dfY,temp);
        dfEst(1) = lo + ((daPointer - floor(daPointer)) * (hi - lo));
        if(dfEst(1) < 0.0)
            dfEst(1) = 0.0;
        end
        if(dfEst(1) > 260.0)
            dfEst(1) = 260.0;
        end

    % and then repeat as required
    %m = 1;  % for if iterations == 1
    if (iterations > 1)
        for m = 2:iterations
            % use improved df estimate to improve da estimate
            dfPointer = 1 + (dfEst(m-1) * dfPointerMult);
            if (dfPointer < 1)
                dfPointer = 1;
            end
            if (dfPointer > 100)
                dfPointer = 100;
            end
            xMax = max(daX(:,ceil(dfPointer)));
            xMin = min(daX(:,ceil(dfPointer)));
            temp = da(ix);
            if (xMax < temp)
                temp = xMax;
            end
            if (temp < xMin)
                temp = xMin;
            end
            hi = interp1DLin(daX(:,ceil(dfPointer)),daY,temp);
            xMax = max(daX(:,floor(dfPointer)));
            xMin = min(daX(:,floor(dfPointer)));
            temp = da(ix);
            if (xMax < temp)
                temp = xMax;
            end
            if (temp < xMin)
                temp = xMin;
            end
            lo = interp1DLin(daX(:,floor(dfPointer)),daY,temp);
            daEst(m) = lo + ((dfPointer - floor(dfPointer)) * (hi - lo));
            if(daEst(m) < 0.0)
                daEst(m) = 0.0;
            end
            if(daEst(m) > 96.0)
                daEst(m) = 96.0;
            end
        
            % now use improved da estimate to improve df estimate
            daPointer = 1 + (daEst(m) * daPointerMult);
            if (daPointer < 1)
                daPointer = 1;
            end
            if (daPointer > 100)
                daPointer = 100;
            end
            xMax = max(dfX(ceil(daPointer),:));
            xMin = min(dfX(ceil(daPointer),:));
            temp = df(ix);
            if (xMax < temp)
                temp = xMax;
            end
            if (temp < xMin)
                temp = xMin;
            end
            hi = interp1DLin(dfX(ceil(daPointer),:),dfY,temp);
            xMax = max(dfX(floor(daPointer),:));
            xMin = min(dfX(floor(daPointer),:));
            temp = df(ix);
            if (xMax < temp)
                temp = xMax;
            end
            if (temp < xMin)
                temp = xMin;
            end
            lo = interp1DLin(dfX(floor(daPointer),:),dfY,temp);
            dfEst(m) = lo + ((daPointer - floor(daPointer)) * (hi - lo));
            if(dfEst(m) < 0.0)
                dfEst(m) = 0.0;
            end
            if(dfEst(m) > 260.0)
                dfEst(m) = 260.0;
            end
        end
    end
    if (iterations>3)
        % take mean of daEst and dfEst 
        df(ix) = mean(dfEst(m-1:m));
        da(ix) = mean(daEst(m-1:m));
    elseif (iterations==1)
        df(ix) = dfEst(m);
        da(ix) = daEst(m);
    end
    
    % chiSquared stuff now
    chiX = interp2DLin((da(ix)*daPointerMult)+1,(df(ix)*dfPointerMult)+1,chiExpect);
    % calculate difference
    chiSquared(ix) = abs(chiSquared(ix) - chiX);
    % finally apply polarity to estimates
    df(ix) = dfPol*df(ix);
    da(ix) = daPol*da(ix);
end
sineIndex = find(chiSquared > 0 & chiSquared < chiSquaredThreshold);
%**************************************************************************
function[amp] = ampCorrection(amp,sineIndex,ampCorrect,dfY,daY,df,da,freq);
global ampScaler;
global daPointerMult;
global dfPointerMult;
for n = 1:length(sineIndex)
    ix = sineIndex(n);
    % ampFactor = interp2(dfY,daY,ampCorrect,abs(df(ix)),abs(da(ix)));
    ampFactor = ampCorrect(round(abs(da(ix))*daPointerMult)+1,round(abs(df(ix))*dfPointerMult)+1);
    % then correct according to window function
    amp(ix) = amp(ix)*ampFactor*ampScaler;
    amp(ix) = windowCorrection(amp(ix),freq(ix),ampFactor);  
end
%**************************************************************************
function[amp] = windowCorrection(peakAmp,freq,ampFactor); 
global Fs;
global FFTSize
global binWidth;
global frameLength;

%calculate deviation from centre of the nearest bin
deviation = freq/binWidth;
deviation = deviation - floor(deviation);
if (deviation > 0.5)
   deviation = 1.0 - deviation;
end
deviation = (deviation * frameLength)/FFTSize;

%scale the amplitude
deviationSquared = deviation^2;
if (deviation == 0.0)
   amp = peakAmp;	%no need to correct estimate
else
   scaler = (sin(pi * deviation))/((pi * deviation) * (1 - deviationSquared));
   amp = peakAmp/(scaler^(1/ampFactor));
end
%**************************************************************************
function[freq] = actualFreqMean(freq,df,da,sineIndex);
% function to calculate actual mean frequency from amplitude weighted
% output of FFT when Hann window is used
for n = 1:length(sineIndex)
    ix = sineIndex(n);
    if (da(ix)~=0)
        ampChange = 10^(da(ix)*0.025);  % da/40
        LA = log(ampChange);
        % calculate mean of amplitude function
        m1 = (ampChange-(1/ampChange));
        m2 = (1/LA)-(LA/((pi^2)+(LA^2)));
        meanAmp = m1*m2;
        meanFreq = meanAmp*freq(ix);
        % calculate df dependent terms
        m1 = (df(ix)*0.5)/(LA^2);
        m2 = (ampChange*(LA-1))+((1/ampChange)*(LA+1));
        i2 = m1*m2;
        m1 = (df(ix)*0.5)/(((pi^2)+(LA^2))^2);
        m2 = -ampChange*((pi^2)+((pi^2)*LA)-(LA^2)+(LA^3));
        m3 = (1/ampChange)*((pi^2)-((pi^2)*LA)-(LA^2)-(LA^3));
        i4 = m1*(m2+m3);
        % calculate actualMeanFreq
        m1 = meanFreq-(i2+i4);
        m2 = 1/(ampChange-(1/ampChange));
        m3 = 1/((1/LA)-(LA/((pi^2)+(LA^2))));
        freq(ix) = m1*m2*m3;
    end
end
%**************************************************************************
function[newSineIndex,correctFreq] = freqTest(freq,correctFreq,sineIndex);
global binWidth
global recipBinWidth
a = 1;
newSineIndex = zeros(0,0);
for n = 1:length(sineIndex)
    ix = sineIndex(n);
    deviation = ceil(abs(correctFreq(ix)-freq(ix))*recipBinWidth);
    if (deviation < 1)
        deviation = 1;
    end
    if(deviation > abs((freq(ix)*recipBinWidth) - (ix-1)))
        newSineIndex(a) = ix;
        a = a+1;
    end
end
%**************************************************************************
function[amp,df,da,endFreq,sineDf,linkIndex,linkDistance,newIndex] = sineLink(sineIndex,freq,amp,df,da,prevEndFreq)
global synProp;
global freqLinkThreshold;
% adjust df and da estimates to fit synthesis frame
sineDf = df(sineIndex); % df values to be copied over for linking in next frame
linkIndex = -1.0*ones(1,length(prevEndFreq));  % stores index of closest link
linkDistance = freqLinkThreshold * ones(1,length(prevEndFreq));   % stores distance (quality) of closest link
newIndex = 1:length(sineIndex);   % pointers to sineIndex which are not linked to previous frames
% find start frequency of each current sinusoid
diff = 0.5*sineDf*synProp;
startFreq = freq(sineIndex) - diff;
endFreq = freq(sineIndex) + diff;
if(length(prevEndFreq)<1)
    return;
end
% find closest links
for a = 1:2 % run at least twice to allow re-assignment of links which have been unassigned
    for n = 1:length(prevEndFreq)
        for m = 1:length(startFreq)
            temp1 = abs(prevEndFreq(n)-startFreq(m));
            if (temp1 < linkDistance(n))
                temp2 = find(linkIndex==m);
                if(temp2)   % has it already been assigned?
                    if(linkDistance(temp2)>temp1)   % assign new link
                        linkDistance(n) = temp1;
                        linkIndex(n) = m;
                        newIndex = newIndex(find(newIndex~=m)); % remove this from 'new' list
                        linkDistance(temp2) = freqLinkThreshold; % unassign previous link
                        newIndex = [newIndex,linkIndex(temp2)]; % put back on the 'new' list
                        linkIndex(temp2) = -1.0;
                    end
                    % otherwise do nothing (i.e. do not assign)
                else    % if it has not been assigned then assign it
                    linkDistance(n) = temp1;
                    linkIndex(n) = m;
                    newIndex = newIndex(find(newIndex~=m)); % remove this from 'new' list
                end
            end
        end
    end
end
%**************************************************************************
function[unLinkedSine,continuedSine,newSine,endAmp,endPhi] = synth(sineIndex,linkIndex,newIndex,prevEndFreq,prevEndAmp,prevEndPhi,prevDf,freq,synAmp,df,da,phi);
global synLength;
global synProp;
global ampFadeMult;
global freqMult;
global frameCount;
global temp;
global standardFFT;
global Fs;
global meanMult;
unLinkedSine = zeros(1,synLength);
continuedSine = zeros(1,synLength);
newSine = zeros(1,synLength);
endAmp = zeros(1,length(sineIndex));
endPhi = zeros(1,length(sineIndex));
endDf = zeros(1,length(sineIndex));
if(linkIndex)
    for n = 1:length(linkIndex)
        if(linkIndex(n)==-1)
            F = prevEndFreq(n)+(prevDf(n)*0.5);
            A = prevEndAmp(n)*ampFadeMult;
            % MATLAB version of function
            %unLinkedSine = unLinkedSine + sineGenerateSP(F,A,prevDf(n)*synProp,-96,synLength,prevEndPhi(n)+(prevEndFreq(n)*freqMult),Fs,meanMult);
            % C version of function
            unLinkedSine = unLinkedSine + CSineGenerateSP(F,A,prevDf(n)*synProp,-96,synLength,prevEndPhi(n)+(prevEndFreq(n)*freqMult),Fs,meanMult);
        else
            ix = linkIndex(n);
            isx = sineIndex(ix);
            % MATLAB version of function
            %[temp,endAmp(ix),endPhi(ix)] = sineGenerateSP(freq(isx),synAmp(isx),df(isx)*synProp,da(isx)*synProp,synLength,prevEndPhi(n)+(prevEndFreq(n)*freqMult),Fs,meanMult);
            % C version of function
            [temp,endAmp(ix),endPhi(ix)] = CSineGenerateSP(freq(isx),synAmp(isx),df(isx)*synProp,da(isx)*synProp,synLength,prevEndPhi(n)+(prevEndFreq(n)*freqMult),Fs,meanMult);
            t1 = synAmp(isx);
            t2 = da(isx);
            t3 = endAmp(ix);
            continuedSine = continuedSine + temp;
        end
    end
end
% synthesize new sines
if(newIndex)
    for n = 1:length(newIndex)
        ix = newIndex(n);
        isx = sineIndex(ix);
        phi = angle(standardFFT(isx))+(pi*0.5);
        if(da(isx)~=0)
            recipMeanAmp = da(isx)/((20/log(10))*(10^(da(isx)*0.025) - 10^(da(isx)*-0.025))); % 0.025 = 1/4;
            midAmp = synAmp(isx) * recipMeanAmp;
            endAmp(ix) = midAmp*10^(da(isx)*0.025*synProp);
        else
            endAmp(ix) = synAmp(isx);
        end
        A = endAmp(ix)*ampFadeMult;
        % MATLAB version of function
        %[temp,endAmp(ix),endPhi(ix)] = sineGenerateCP(freq(isx),A,df(isx)*synProp,96,synLength,phi,Fs,meanMult);
        % C version of function
        [temp,endAmp(ix),endPhi(ix)] = CSineGenerateCP(freq(isx),A,df(isx)*synProp,96,synLength,phi,Fs,meanMult);
        newSine = newSine + temp;
    end
end
%**************************************************************************
function[sinusoid,endAmp,endPhi] = sineGenerateSP(f,a,df,da,N,startPhase,Fs,meanMult);
% Returns a single frame of a sinusoid with characteristics:
% f: mid-point frequency (Hz)
% a: mid-point amplitude
% df: frequency change within frame (Hz)
% da: amplitude change within frame (dB)
% N is frame size
% startPhase is phase of first sample

doublePi = 2 * pi;
Mlt = doublePi/Fs;

if (df == 0)
   % generate phases for stationary sinusoid
    if (f ~= 0)
        phases = startPhase: ((f/Fs) * doublePi): ((N-1) * ((f/Fs) * doublePi) + startPhase);
    else
        % generate vector of identical phase values
        phases = startPhase.*ones(1,N);
    end
else
   % generate phases for sinusoid with frequency ramp
   phases = linspace(f-0.5*df,f+0.5*df,N);
   phases = cumsum(phases.*Mlt);
   phases = phases - phases(1) + startPhase;
end
% generate sinusoid
sinusoid = sin(phases);
endPhi = phases(end);

if(da ~= 0.0)
   % convert from dB
   ampChange = 10^(da*0.0250);
   % apply exponential amplitude modulation to sinusoid
   ampRamp = logspace(log10(1/ampChange),log10(ampChange), N);
   % scale overall amplitude so that a is still the mean amplitude
   m = meanMult*(ampChange - (1/ampChange))/da;
   ampRamp = (a/m) * ampRamp;
   sinusoid = sinusoid .* ampRamp; 
   endAmp = ampRamp(end);
else
   sinusoid = sinusoid * a;
   endAmp = a;
end
%**************************************************************************
function[sinusoid,endAmp,endPhi] = sineGenerateCP(f,a,df,da,N,centrePhase,Fs,meanMult);
% Returns a single frame of a sinusoid with characteristics:
% f: mid-point frequency (Hz)
% a: mid-point amplitude
% df: frequency change within frame (Hz)
% da: amplitude change within frame (dB)
% N is frame size
% centrePhase is phase of middle sample

doublePi = 2 * pi;
Mlt = doublePi/Fs;

if (df == 0)
   % generate phases for stationary sinusoid
    if (f ~= 0)
        phases = centrePhase: ((f/Fs) * doublePi): ((N-1) * ((f/Fs) * doublePi) + centrePhase);
    else
        % generate vector of identical phase values
        phases = centrePhase.*ones(1,N);
    end
else
   % generate phases for sinusoid with frequency ramp
   phases = linspace(f-0.5*df,f+0.5*df,N);
   phases = cumsum(phases.*Mlt);
   phases = phases - phases(1) + centrePhase;
end
% shift phases so that centrePhase is at centre sample
phases = phases - (phases(ceil(N/2))-centrePhase);
% generate sinusoid
sinusoid = sin(phases);
endPhi = phases(end);

if(da ~= 0.0)
   % convert from dB
   ampChange = 10^(da*0.025);
   % apply exponential amplitude modulation to sinusoid
   ampRamp = logspace(log10(1/ampChange),log10(ampChange), N);
   % scale overall amplitude so that a is still the mean amplitude
   m = meanMult*(ampChange - (1/ampChange))/da;
   ampRamp = (a/m) * ampRamp;
   sinusoid = sinusoid .* ampRamp; 
   endAmp = ampRamp(end);
else
   sinusoid = sinusoid * a;
   endAmp = a;
end
%**************************************************************************
function[standardFFT] = spectralSubtract(standardFFT,sineIndex,residualRange,complexSize,recipBinWidth,df);
for n = 1:length(sineIndex)
    ix = sineIndex(n);
    temp = abs(round(df(ix)*recipBinWidth*0.5)); % account for df
    lower = ix-residualRange-temp;
    upper = ix+residualRange+temp;
    if (upper >complexSize)
        upper = complexSize;
    end
    if (lower<1)
        lower = 1;
    end
    standardFFT(lower:upper) = 0;
end
%**************************************************************************
function[decMag,decFreq,decBW,undecMag,undecFreq,undecBW] = AnalyticWaveletCombSplit(standardFFT,tempFFT,FFTSize,complexSize,frameLength,num_undec,num_dec,INIT,fL_5,decFreqMult,Fs2Pi,Fs,Fs2Powers,decMag,decFreq,decBW,undecMag,undecFreq,undecBW,scaleFreqRange,splitFreqLimit,centreScaleFreqs);
% Partially Decimated B-Spline Split Wavelet Transform

% multiply residual by wavelet initialisation filter and calculate discrete
% time analytic signal
tempFFT(2:complexSize-1) = 2.*standardFFT(2:complexSize-1).*INIT(2:complexSize-1);
tempFFT(complexSize+1:end) = 0;
tempFFT([1,complexSize]) = standardFFT([1,complexSize]).*INIT([1,complexSize]);
temp = ifft(tempFFT);
analytic(fL_5+1:frameLength) = temp(1:fL_5+1);
analytic(1:fL_5) = temp(FFTSize-fL_5+1:FFTSize);
    
% undecimated
if (num_undec>0)
    [undecRealDetail,undecRealLower,undecRealUpper,undecRealOutput] = ShensaUDWTSplit(real(analytic));
    [undecImagDetail,undecImagLower,undecImagUpper,undecImagOutput] = ShensaUDWTSplit(imag(analytic));  
end

% decimated
if (num_dec>0)
    if (num_undec>0)
        [decRealDetail,decRealLower,decRealUpper] = ShensaUDWTDecSplit(undecRealOutput(1:2:end));
        [decImagDetail,decImagLower,decImagUpper] = ShensaUDWTDecSplit(undecImagOutput(1:2:end));
    else
        [decRealDetail,decRealLower,decRealUpper]  = ShensaUDWTDecSplit(real(analytic));
        [decImagDetail,decImagLower,decImagUpper] = ShensaUDWTDecSplit(imag(analytic));
    end
end

% calculate mag and frequency data
if (num_dec~=0)
    for n = 1:num_dec
        npnu = n + num_undec;
        decMag{n} = ((decRealDetail{n}.^2) + (decImagDetail{n}.^2)).^0.5; 
        if (npnu<9)
            % MATLAB version of function
            %decFreq{n} = diff(unwrap(atan2(decImagDetail{n},decRealDetail{n})))*Fs2Pi*decFreqMult(n); 
            % MEX version of function
            decFreq{n} = diff(CPhaseUnwrap(decImagDetail{n},decRealDetail{n}))*Fs2Pi*decFreqMult(n); 
            decFreq{n}(find(decFreq{n}<0)) = decFreq{n}(find(decFreq{n}<0))+Fs2Powers(npnu);
        else
            decFreq{n} = ones(1,length(decMag{n})-1)*centreScaleFreqs(npnu);
        end
        decMag{n}(find(decFreq{n}>scaleFreqRange(npnu)|decFreq{n}<scaleFreqRange(npnu+1)))= 0;   % zero magnitudes outside of range 
        decFreq{n}(find(decFreq{n}>scaleFreqRange(npnu)|decFreq{n}<scaleFreqRange(npnu+1))) = centreScaleFreqs(npnu); % set freq in middle of range
        if (npnu<9)
            % MATLAB version of function
            %decBW{n} = abs(diff(unwrap(atan2(decImagUpper{n},decRealUpper{n}))) - diff(unwrap(atan2(decImagLower{n},decRealLower{n}))))*Fs2Pi*decFreqMult(n);
            % MEX version of function
            decBW{n} = abs(diff(CPhaseUnwrap(decImagUpper{n},decRealUpper{n})) - diff(CPhaseUnwrap(decImagLower{n},decRealLower{n})))*Fs2Pi*decFreqMult(n);
            decBW{n}(find(decBW{n}>splitFreqLimit(npnu))) = splitFreqLimit(npnu);
        else
            decBW{n} = ones(1,length(decMag{n})-1)*splitFreqLimit(npnu);
        end
        decMag{n}(1:end-1) = decMag{n}(1:end-1)*sqrt(splitFreqLimit(npnu))./sqrt(decBW{n});
    end
end
if (num_undec~=0)
    undecMag = ((undecRealDetail.^2) + (undecImagDetail.^2)).^0.5; 
    for n = 1:num_undec
        if (n<9)
            % MATLAB version of function
            %undecFreq(n,:) = diff(unwrap(atan2(undecImagDetail(n,:),undecRealDetail(n,:))))*Fs2Pi; 
            % MEX version of function
            undecFreq(n,:) = diff(CPhaseUnwrap(undecImagDetail(n,:),undecRealDetail(n,:)))*Fs2Pi; 
            undecFreq(n,find(undecFreq(n,:)<0)) = undecFreq(n,find(undecFreq(n,:)<0))+ Fs2Powers(n);
        else
            undecFreq(n,:) = 0.5*(scaleFreqRange(n+1)+scaleFreqRange(n));
        end
        undecMag(n,find(undecFreq(n,:)>scaleFreqRange(n)|undecFreq(n,:)<scaleFreqRange(n+1))) = 0;   % zero magnitudes outside of range
        undecFreq(n,find(undecFreq(n,:)>scaleFreqRange(n)|undecFreq(n,:)<scaleFreqRange(n+1))) = 0.5*(scaleFreqRange(n+1)+scaleFreqRange(n)); % set freq in middle of range
        if (n<9)
            % MATLAB version of function
            %undecBW(n,:) = abs(diff(unwrap(atan2(undecImagUpper(n,:),undecRealUpper(n,:)))) - diff(unwrap(atan2(undecImagLower(n,:),undecRealLower(n,:)))))*Fs2Pi; 
            % MEX version of function
            undecBW(n,:) = abs(diff(CPhaseUnwrap(undecImagUpper(n,:),undecRealUpper(n,:))) - diff(CPhaseUnwrap(undecImagLower(n,:),undecRealLower(n,:))))*Fs2Pi; 
            undecBW(n,find(undecBW(n,:)>splitFreqLimit(n))) = splitFreqLimit(n);
        else
            undecBW(n,:) = splitFreqLimit(n);
        end
        undecMag(n,1:end-1) = undecMag(n,1:end-1)*sqrt(splitFreqLimit(n))/sqrt(undecBW(n,:));  
    end
end
%**************************************************************************
function[sFilter,wFilter] = BSplineMake(ord)
% function to produce low and high pass
% filter coefficients for the B-spline
% biorthogonal wavelet. The ord is the 
% order of the spline function used (e.g.
% ord 3 is a cubic spline)
% init is the  
% initialisation filter (Unser eqn 3.32)

% sclaing function (sFilter)
% binomial kernel of order n
% see Unser "Polynomial Spline etc." eqn 2.4
% and Wang\Chui "On compactly supported spline.." eqn 1.9
mult = 1/(2^ord);
for n = 1:(ord+2)
    sFilter(n) = nchoosek(ord+1,n-1)*mult;
end
% find 2n+1 spline
order2n1 = (2*ord)+1;
x = [fix(-order2n1/2):fix(order2n1/2)];
% eqn 10 from "Splines, A Perfect Fit" Unser
% calculate one half 
b = BSpline(order2n1,-abs(x));
% produce wavelet filter (wFilter)
LPFmodulator = (ones(size(sFilter))).*-1;
if (rem(ord,2)~=0)
    for n = 1:length(LPFmodulator)
        LPFmodulator(n) = LPFmodulator(n)^(n);
    end
else
    for n = 1:length(LPFmodulator)
        LPFmodulator(n) = LPFmodulator(n)^n;
    end
end
bModulator = (ones(size(b))).*-1;
if (rem(ord,2)~=0)
    for n = 1:length(bModulator)
        bModulator(n) = bModulator(n)^n;
    end
else
    for n = 1:length(bModulator)
        bModulator(n) = bModulator(n)^n;
    end
end
wFilter = CConv((sFilter.*LPFmodulator),(b.*bModulator));
% make sure centre value is positive
if (rem(ord,2)~=0)
    if (wFilter(ceil(length(wFilter)/2))<0)
        wFilter = wFilter.*-1;
    end
end
%**************************************************************************
function[B] = BSpline(ord,x);
o1 = ord+1;
mult = 1/factorial(ord);
for n= 1:length(x)
    B(n) = 0;
    for j = 0:o1
        a = (-1)^j;
        b = nchoosek(o1,j);
        c = x(n)+(o1/2)-j;
        if (c<0)
            B(n) = B(n);
        else
            B(n) = B(n)+(a*b*(c^ord));
        end
    end
end
B = B.*mult;
%**************************************************************************
function[detail,upper,lower,signal] = ShensaUDWTSplit(signal)
% function[detail,approx,upper,lower] = ShensaUDWTSplit(signal)
% implementation of undecimated wavelet transform
% proposed in Shensa "Wedding the A Trous and Mallat Algorithms" 
% An alternative but numerically equivalent to the A Trous
% algorithm
% signal: input
% LPF: low pass filter (scaling)
% HPF: high pass filter (wavelet)
% The 'split' part refers to the splitting trick (splitting the detail
% at each level in to two parts). The purpose of this is to provide 
% more frequency localised filters from which the 'width' of the 
% detail signal can be derived.
global LPF;
global HPF;
global lH;
global lL;
global frameLength;
global num_undec;
global synLength;
len = length(signal);

det{1} = CConv(HPF,signal);
app{1} = CConv(LPF,signal);  

for n = 2:num_undec
    tempA = app{n-1};
    tempD = det{n-1};
    tempDetail = zeros(1,(2^(n-1))*(lH-1)+length(tempA));   %pre-allocate memory
    tempApprox = zeros(1,(2^(n-1))*(lL-1)+length(tempA));
    tempUpper = zeros(1,(2^(n-1))*(lH-1)+length(tempD));
    tempLower = zeros(1,(2^(n-1))*(lL-1)+length(tempD));
    M = 2^(n-1);
    for m = 1:M
        tempDetail(m:M:(M*(lH-1))+length(tempA))=CConv(HPF,tempA(m:M:end));
        tempApprox(m:M:(M*(lL-1))+length(tempA))=CConv(LPF,tempA(m:M:end));
        tempUpper(m:M:(M*(lH-1))+length(tempD))=CConv(HPF,tempD(m:M:end));
        tempLower(m:M:(M*(lL-1))+length(tempD))=CConv(LPF,tempD(m:M:end));
    end
    det{n} = tempDetail;
    app{n} = tempApprox;
    upp{n-1} = tempUpper;
    low{n-1} = tempLower;
end
% get final level of 'splits'
if (num_undec==1)
    n=1;
end
tempDetail = det{n};

tempUpper = zeros(1,(2^n)*(lH-1)+length(tempDetail));   %pre-allocate memory
tempLower = zeros(1,(2^n)*(lL-1)+length(tempDetail));
for m = 1:2^n
    tempUpper(m:2^n:((2^n)*(lH-1))+length(tempDetail))=CConv(HPF,tempDetail(m:2^n:end));
    tempLower(m:2^n:((2^n)*(lL-1))+length(tempDetail))=CConv(LPF,tempDetail(m:2^n:end));
end

upp{n} = tempUpper;
low{n} = tempLower;
signal = app{n};

for n = 1:length(app)
    tempD = det{n};
    tempL = low{n};
    tempU = upp{n};
    lowerD = ceil((length(tempD)-synLength)*0.5);
    lowerL = ceil((length(tempL)-synLength)*0.5);
    lowerU = ceil((length(tempU)-synLength)*0.5);
    upperD = lowerD+synLength;
    upperL = lowerL+synLength;
    upperU = lowerU+synLength;
    detail(n,:) = tempD(lowerD:upperD);
    lower(n,:) = tempL(lowerL:upperL);
    upper(n,:) = tempU(lowerU:upperU);
end
%**************************************************************************
function[detail,upper,lower] = ShensaUDWTDecSplit(app)
% function = ShensaUDWT(signal,LPF,HPF)
% implementation of undecimated wavelet transform
% proposed in Shensa "Wedding the A Trous and Mallat Algorithms" .
% An alternative, but computationally and numerically equivalent, to the A Trous
% algorithm
% This version actually decimates the input at each iteration
% and is intended to be used as the decimated part of the partially
% decimated transform
% app: input approximation 
global LPF;
global HPF;
global lL;
global lH;
global num_dec;
global num_undec;
global synLength;
global p;   % number of holes in filters
global p2;  % number of holes in split filters
global decLen; % lengths of output sequences

offset = abs(rem(lH,2)-1);  %for even order splines
M = ceil(2^(num_undec-1));

for n = 1:num_dec
    len = length(app);
    tempDetail = zeros(1,p*(lH-1)+len); %pre-allocate memory
    tempApprox = zeros(1,p*(lL-1)+len);
    for m = 1:M
        tempDetail(m:p:p*(lH-1)+len)=CConv(HPF,app(m:p:end));
        tempApprox(m:p:p*(lL-1)+len)=CConv(LPF,app(m:p:end));
    end
    len = length(tempDetail);
    ixL = ceil((len-decLen(n))*0.5);
    ixU = ixL+decLen(n);
    detail{n}=tempDetail(ixL:ixU);
    app = tempApprox(1:2:end);
    tempUpper = zeros(1,2*p*(lH-1)+len);  %pre-allocate memory
    tempLower = zeros(1,2*p*(lL-1)+len);
    for m = 1:M*2 
        tempUpper(m:p2:p2*(lH-1)+len)=CConv(HPF,tempDetail(m:p2:end));
        tempLower(m:p2:p2*(lL-1)+len)=CConv(LPF,tempDetail(m:p2:end));
    end
    ixL = ceil((length(tempLower)-decLen(n))*0.5);
    ixH = ixL+decLen(n);
    lower{n} = tempLower(ixL:ixH);
    ixL = ceil((length(tempUpper)-decLen(n))*0.5);
    ixH = ixL+decLen(n);
    upper{n} = tempUpper(ixL:ixH);
end
%**************************************************************************
function[] = EQ();
global noise;
global decMag;
global decFreq;
global decBW;
global undecMag;
global undecFreq;
global undecBW;
global num_dec;
global num_undec;
global xnm1;
global xnm2;
global ynm1;
global ynm2;
global temp;
global noiseOut;
global synLength;
global sqrt0_5;
global freqMult;
global dB96Mult;
global Fs;
global decLen;
global decSamples;
global G0;
global G;
global GB;
global G2;
global GB2;
global G02;
global G2_GB2;
global GB2_G02;
global GB2G02G2;
global PI2;
global G2PI2;
global G2_G02;
global tanTable;
global tanLookupMult;
p = 0.5.^[0:7,7.6503,7.7795];
BWGain = sqrt0_5*dB96Mult;

noiseOut  = zeros(1,synLength); % clear out from previous time around

for n = 1:num_undec
    for m = 1:synLength
        % calculate coefficients
        filterCoeff = OrfanidisCoeffCalc(1,dB96Mult,sqrt0_5*dB96Mult,undecFreq(n,m)*freqMult,undecBW(n,m)*freqMult);
        temp(m) = sum(filterCoeff.*[undecMag(n,m)*p(n)*noise(m),xnm1(n),xnm2(n),-1*ynm1(n),-1*ynm2(n)]);
        xnm2(n) = xnm1(n);
        xnm1(n) = undecMag(n,m)*p(n)*noise(m);
        ynm2(n) = ynm1(n);
        ynm1(n) = temp(m);
    end
    noiseOut = noiseOut + temp;
end

for n = 1:num_dec
    npnu = n + num_undec;
    for m = 1:decLen(n)
        % only calculate coefficients when parameters are updated
        filterCoeff = OrfanidisCoeffCalc(1,dB96Mult,sqrt0_5*dB96Mult,decFreq{n}(m)*freqMult,decBW{n}(m)*freqMult);
        dM = decMag{n}(m)*p(npnu);
        idxOffset = (m-1)*decSamples(n);
        for k = 1:decSamples(n)  
            temp(idxOffset+k) = sum(filterCoeff.*[dM*noise(idxOffset+k),xnm1(npnu),xnm2(npnu),-1*ynm1(npnu),-1*ynm2(npnu)]);
            xnm2(npnu) = xnm1(npnu);
            xnm1(npnu) = dM*noise(idxOffset+k);
            ynm2(npnu) = ynm1(npnu);
            ynm1(npnu) = temp(idxOffset+k);
        end
    end
    noiseOut = noiseOut + temp;
end
%**************************************************************************
function[ba] = BristowCoeffCalc(G0,G,GB,W0,DW)
beta = sqrt((GB^2-G0^2)/(G^2-GB^2))*tan(DW*0.5);
CS = 2*cos(W0);
ba = [(G0+(G*beta)),-(G0*CS),(G0-(G*beta)),-(G0*CS),(1-beta)]/(1+beta);
%**************************************************************************
function[ba] = OrfanidisCoeffCalc(G0,G,GB,W0,DW)
G2 = G^2;
GB2 = GB^2;
G02 = G0^2;
G2_GB2 = G2-GB2;
GB2_G02 = GB2-G02;
GB2G02G2 = GB2_G02/G2_GB2;
PI2 = pi^2;
G2PI2 = G2*PI2;
G2_G02 = G2-G02;

t = (G02*((W0^2-PI2)^2))+(G2*PI2*DW^2*GB2G02G2);
b = ((W0^2-PI2)^2)+ (PI2*DW^2*GB2G02G2);
G12 = t/b;
G1 = sqrt(G12);
W0 = tan(W0*0.5);
W2 = sqrt((G2-G12)/G2_G02)*W0^2;
DW = (1 + (sqrt(GB2_G02/abs(GB2-G12))*W2))*tan(DW*0.5);
C = ((DW^2)*abs(GB2-G12))-(2*W2*(abs(GB2-(G0*G1))-sqrt(GB2_G02*abs(GB2-G12))));
D = 2*W2*(abs(G2-(G0*G1))-sqrt(G2_G02*(G2-G12)));
E = sqrt((C+D)/G2_GB2);
F = sqrt((G2*C+GB2*D)/G2_GB2);  
ba = [(G1+(G0*W2)+F),-2*(G1-(G0*W2)),(G1+(G0*W2)-F),-2*(1-W2),(1+W2-E)]/(1+W2+E);
%**************************************************************************
function[yi] = interp1DLin(x,Y,xi)
% alternative to interp1
if (xi>=x(end))
    yi = Y(end);
elseif (xi<=x(1))
    yi = Y(1);
else
    idx = find(x>xi,1);
    a = x(idx);
    b = x(idx-1);
    frac = (xi-b)/(a-b);
    yi = Y(idx-1)+(frac*(Y(idx)-Y(idx-1)));
end
%**************************************************************************
function[ci] = interp2DLin(xi,yi,table)
% function[ci] = interp2DLin(xi,yi,table)
% alternative to interp2

% determine 4 coordinates (2 x values, 2 y values)
[m,n] = size(table);
if (xi>=m)
    xi = m;
    x1 = m-1;
    x2 = m;
elseif (xi<=1)
    xi = 1;
    x1 = 1;
    x2 = 2;
else
    x1 = floor(xi);
    x2 = ceil(xi);
end

if (yi>=n)
    yi = n;
    y1 = n-1;
    y2 = n;
elseif (yi<=1)
    yi = 1;
    y1 = 1;
    y2 = 2;
else
    y1 = floor(yi);
    y2 = ceil(yi);
end
% find wieghted points from coordinates
p11 = table(x1,y1)*(x2-xi)*(y2-yi);
p21 = table(x2,y1)*(xi-x1)*(y2-yi);
p12 = table(x1,y2)*(x2-xi)*(yi-y1);
p22 = table(x2,y2)*(xi-x1)*(yi-y1);

ci = p11+p21+p12+p22;
%**************************************************************************
function[decMag,undecMag] = waveletMagCorrect(num_dec,num_undec,decFreq,decMag,undecFreq,undecMag,centreFreqMag,recipFs,order);
% function to correct magnitude given deviation from wavelet filter peak 
for n = 1:num_undec
    undecFreq(n,:) = undecFreq(n,:).*recipFs.*(2^(n-1));
    Cf = (sin((2*pi*undecFreq(n,:))).^2./(undecFreq(n,:).*(undecFreq(n,:)-0.5).^2)).^(order+1);
    undecMag(n,1:end-1) = undecMag(n,1:end-1).*(centreFreqMag./Cf);
end

for n = 1:num_dec
    clear Cf;
    decFreq{n} = decFreq{n}.*recipFs.*(2^(n+num_undec-1));
    Cf = (sin((2*pi*decFreq{n})).^2./(decFreq{n}.*(decFreq{n}-0.5).^2)).^(order+1);
    decMag{n}(1:end-1) = decMag{n}(1:end-1).*(centreFreqMag./Cf);
end
    